# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# This program uses Apache Spark to generate example binary Variant data
#
# Requirements
# pip install pyarrow
# pip install pyspark
#
# Last run with Spark 4.0 preview 2:
# https://spark.apache.org/news/spark-4.0.0-preview2.html

from pyspark.sql import SparkSession
import pyarrow.parquet as pq
import os
import json

# Initialize Spark session and create variant data via SQL
spark = SparkSession.builder \
    .appName("PySpark SQL Example") \
    .getOrCreate()

# recursively cleanup the spark-warehouse directory
if os.path.exists('spark-warehouse'):
    for root, dirs, files in os.walk('spark-warehouse', topdown=False):
        for name in files:
            os.remove(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name))


# Create a table with variant and insert various types into it
#
# This writes data files into spark-warehouse/output
sql = """
CREATE TABLE T (name VARCHAR(2000), variant_col VARIANT);

-------------------------------
-- Primitive type (basic_type=0)
-------------------------------
-- One row with a value from each type listed in 
-- https://github.com/apache/parquet-format/blob/master/VariantEncoding.md#encoding-types
--
-- Spark Types: https://spark.apache.org/docs/latest/sql-ref-datatypes.html
-- Note: must use explicit typecasts as Spark returns an error for implicit casts
INSERT INTO T VALUES ('primitive_null', NULL);
INSERT INTO T VALUES ('primitive_boolean_true', true::Variant);
INSERT INTO T VALUES ('primitive_boolean_false', false::Variant);
INSERT INTO T VALUES ('primitive_int8', 42::Byte::Variant);
INSERT INTO T VALUES ('primitive_int16', 1234::Short::Variant);
INSERT INTO T VALUES ('primitive_int32', 123456::Integer::Variant);
INSERT INTO T VALUES ('primitive_int64', 1234567890123456789::Long::Variant); -- must be too large to fit in int32: https://github.com/apache/parquet-testing/issues/82
INSERT INTO T VALUES ('primitive_double', 1234567890.1234::Double::Variant);
INSERT INTO T VALUES ('primitive_decimal4', 12.34::Decimal(8,2)::Variant);
INSERT INTO T VALUES ('primitive_decimal8', 12345678.90::Decimal(12,2)::Variant);
INSERT INTO T VALUES ('primitive_decimal16', 12345678912345678.90::Decimal(30,2)::Variant);
INSERT INTO T VALUES ('primitive_date', '2025-04-16'::Date::Variant);
INSERT INTO T VALUES ('primitive_timestamp', '2025-04-16T12:34:56.78'::Timestamp::Variant);
INSERT INTO T VALUES ('primitive_timestampntz', '2025-04-16T12:34:56.78'::Timestamp_NTZ::Variant);
INSERT INTO T VALUES ('primitive_float', 1234567890.1234::Float::Variant);
INSERT INTO T VALUES ('primitive_binary', X'31337deadbeefcafe'::Variant);
INSERT INTO T VALUES ('primitive_string', 'This string is longer than 64 bytes and therefore does not fit in a short_string and it also includes several non ascii characters such as 🐢, 💖, ♥️, 🎣 and 🤦!!'::Variant);

-- binary artifacts of 'TimeNTZ'/'timestamp with timezone (NANOS)'/'timestamp without time zone (NANOS)'/'UUID' was generated by the iceberg test code, please ref to https://github.com/apache/parquet-testing/pull/92 for more detail

-------------------------------
-- Short string (basic_type=1)
-------------------------------
INSERT INTO T VALUES ('short_string', 'Less than 64 bytes (❤️ with utf8)'::Variant);

-------------------------------
-- Object (basic_type=2)
-------------------------------
-- Use parse_json to create Variant, as spark does not seem to support casting structs --> Variant.
INSERT INTO T VALUES ('object_empty', parse_json('{}')::Variant);
INSERT INTO T VALUES ('object_primitive', parse_json('{"int_field" : 1, "double_field": 1.23456789, "boolean_true_field": true, "boolean_false_field": false, "string_field": "Apache Parquet", "null_field": null, "timestamp_field": "2025-04-16T12:34:56.78"}')::Variant);
INSERT INTO T VALUES ('object_nested', parse_json('{ "id" : 1, "species" : { "name": "lava monster", "population": 6789}, "observation" : { "time": "12:34:56", "location": "In the Volcano", "value" : { "temperature": 123, "humidity": 456 } } }')::Variant);

-- https://github.com/apache/parquet-testing/issues/77
-- TODO create example variant objects with fields that non-json types (like timestamp, date, etc)
-- Casting from "STRUCT<...>" to "VARIANT"" is not yet supported
-- INSERT INTO T VALUES ('object_primitive', struct(1234.56::Double as double_field, true as boolean_true_field, false as boolean_false_field, '2025-04-16T12:34:56.78'::Timestamp as timestamp_field, 'Apache Parquet' as string_field, null as null_field)::Variant);
--TODO objects with more than 2**8 distinct fields (that require using more than one byte for field offset)
--TODO objects with more than 2**16 distinct fields (that require using more than 2 bytes for field offset)
--TODO objects with more than 2**24 distinct fields (that require using more than 3 bytes for field offset)

-------------------------------
-- Array (basic_type=3)
-------------------------------
INSERT INTO T VALUES ('array_empty', parse_json('[]')::Variant);
INSERT INTO T VALUES ('array_primitive', parse_json('[2, 1, 5, 9]')::Variant);
INSERT INTO T VALUES ('array_nested', parse_json('[ { "id": 1, "thing": { "names": ["Contrarian", "Spider"] } }, null, { "id": 2, "type": "if", "names": ["Apple", "Ray", null] } ]')::Variant);

-- https://github.com/apache/parquet-testing/issues/78
-- TODO arrays with more than 2**8 distinct elements (that require using more than one byte for count)
-- TODO arrays where the total length of all values is greater than 2**8, 2**16, and 2**24 bytes (that require using more than one byte for the offsets)

-------------------------------
-- Output the value to a new table that also has the JSON representation of the variant column
-------------------------------
DROP TABLE IF EXISTS output;
CREATE TABLE output AS SELECT name, variant_col, to_json(variant_col) as json_col FROM T;
"""
for statement in sql.split("\n"):
    statement = statement.strip()
    if not statement or statement.startswith("--"):
        continue
    print("Running SQL:", statement)
    spark.sql(statement)

mypath = 'spark-warehouse/output'
parquet_files = [f for f in os.listdir(mypath) if f.endswith('.parquet')]

# extract the values from the parquet files
data_dictionary = {}
for f in parquet_files:
    table = pq.read_table(os.path.join(mypath, f))
    for row in range(len(table)):
        name = table[0][row]
        # variants are stored as StructArrays with two fields:
        # metadata, and value
        variant_col = table[1][row]
        metadata = variant_col['metadata']
        value = variant_col['value']
        json_value = table[2][row]

        print("Writing metadata for", name)

        # write the metadata, value, and json representation to files
        with open(f"{name}.metadata", "wb") as f:
            buffer = metadata.as_buffer()
            if buffer is not None:
                f.write(buffer)
        with open(f"{name}.value", "wb") as f:
            buffer = value.as_buffer()
            if buffer is not None:
                f.write(buffer)

        # Add the JSON representation to the data dictionary
        name = name.as_py()
        json_value = json_value.as_py()

        if json_value is not None:
            data_dictionary[name] = json.loads(json_value)
        else:
            data_dictionary[name] = None

with open(f"data_dictionary.json", "w") as f:
    f.write(json.dumps(data_dictionary, sort_keys = True, indent=4))

# Note: It is possible to write the output to a single parquet file, using a command
# such as:
# spark.sql("SELECT * FROM output").repartition(1).write.parquet('variant.parquet')
# At the time of writing, this file does not have the logical type annotation for VARIANT



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json

from pathlib import Path

import pyarrow as pa
from pyarrow import parquet
import pyproj
import shapely

HERE = Path(__file__).parent


# Using Wyoming because it is the easiest state to inline into a Python file
WYOMING_LOWRES = (
    "POLYGON ((-111.0 45.0, -111.0 41.0, -104.0 41.0, -104.0 45.0, -111.0 45.0))"
)

# We densify the edges such that there is a point every 0.1 degrees to minimize
# the effect of the edge algorithm and coordinate transformation.
WYOMING_HIRES = shapely.from_wkt(WYOMING_LOWRES).segmentize(0.1).wkt


class WkbType(pa.ExtensionType):
    """Minimal geoarrow.wkb implementation"""

    def __init__(self, crs=None, edges=None, *, storage_type=pa.binary(), **kwargs):
        self.crs = crs
        self.edges = edges
        super().__init__(storage_type, "geoarrow.wkb")

    def __arrow_ext_serialize__(self):
        obj = {"crs": self.crs, "edges": self.edges}
        return json.dumps({k: v for k, v in obj.items() if v}).encode()

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        obj: dict = json.loads(serialized)
        return WkbType(**obj, storage_type=storage_type)


pa.register_extension_type(WkbType())


def write_crs(type, geometry, name, col_name="geometry", metadata=None):
    schema = pa.schema({"wkt": pa.utf8(), col_name: type})

    with parquet.ParquetWriter(
        HERE / name,
        schema,
        # Not sure if there's a way to write metadata without
        # storing the Arrow schema
        store_schema=metadata is not None,
        compression="none",
    ) as writer:
        batch = pa.record_batch(
            {
                "wkt": [geometry.wkt],
                col_name: type.wrap_array(pa.array([geometry.wkb])),
            }
        )
        writer.write_batch(batch)

        if metadata is not None:
            writer.add_key_value_metadata(metadata)


def write_crs_files():
    # Create the Shapely geometry
    geometry = shapely.from_wkt(WYOMING_HIRES)

    # A general purpose coordinate system for the United States
    crs_not_lonlat = pyproj.CRS("EPSG:5070")
    transformer = pyproj.Transformer.from_crs(
        "OGC:CRS84", crs_not_lonlat, always_xy=True
    )
    geometry_not_lonlat = shapely.transform(
        geometry, transformer.transform, interleaved=False
    )

    # Write with the default CRS (i.e., lon/lat)
    write_crs(WkbType(), geometry, "crs-default.parquet")

    # Write a Geography column with the default CRS
    write_crs(
        WkbType(edges="spherical"),
        geometry,
        "crs-geography.parquet",
        col_name="geography",
    )

    # Write a file with the projjson format in the specification
    # and the appropriate metadata key
    write_crs(
        WkbType(crs="projjson:projjson_epsg_5070"),
        geometry_not_lonlat,
        "crs-projjson.parquet",
        metadata={"projjson_epsg_5070": crs_not_lonlat.to_json()},
    )

    # Write a file with the srid format in the specification
    write_crs(WkbType(crs="srid:5070"), geometry_not_lonlat, "crs-srid.parquet")

    # Write a file with an arbitrary value (theoretically allowed by the format
    # and consumers may choose to error or attempt to interpret the value)
    write_crs(
        WkbType(crs=crs_not_lonlat.to_json_dict()),
        geometry_not_lonlat,
        "crs-arbitrary-value.parquet",
    )


def check_crs_schema(name, expected_col_type):
    file = parquet.ParquetFile(HERE / name)

    col = file.schema.column(1)
    col_dict = json.loads(col.logical_type.to_json())
    col_type = col_dict["Type"]
    if col_type != expected_col_type:
        raise ValueError(
            f"Expected '{expected_col_type}' logical type but got '{col_type}'"
        )


def check_crs_crs(name, expected_crs):
    expected_crs = pyproj.CRS(expected_crs)

    file = parquet.ParquetFile(HERE / name, arrow_extensions_enabled=True)
    ext_type = file.schema_arrow.field(1).type
    actual_crs = pyproj.CRS(ext_type.crs)
    if actual_crs != expected_crs:
        raise ValueError(f"Expected '{expected_crs}' crs but got '{actual_crs}'")


def check_crs(name, expected_col_type, expected_crs):
    check_crs_schema(name, expected_col_type)
    check_crs_crs(name, expected_crs)


if __name__ == "__main__":
    write_crs_files()

    check_crs("crs-default.parquet", "Geometry", "OGC:CRS84")
    check_crs("crs-geography.parquet", "Geography", "OGC:CRS84")
    check_crs("crs-projjson.parquet", "Geometry", "EPSG:5070")
    check_crs("crs-srid.parquet", "Geometry", "EPSG:5070")
    check_crs("crs-arbitrary-value.parquet", "Geometry", "EPSG:5070")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json
from pathlib import Path
import re

import geoarrow.pyarrow as ga
import numpy as np
import pyarrow as pa
from pyarrow import parquet
import shapely
import shapely.testing
import yaml

# M value support was added in Shapely 2.1.0
if tuple(shapely.__version__.split(".")) < ("2", "1", "0"):
    raise ImportError("shapely >= 2.1.0 is required")

HERE = Path(__file__).parent

# Shapely doesn't propagate the geometry type for MULTI* Z/M/ZM, so
# we have to write our own geometry type detector to check statistics output
GEOMETRY_TYPE_CODE = {
    "POINT": 1,
    "LINESTRING": 2,
    "POLYGON": 3,
    "MULTIPOINT": 4,
    "MULTILINESTRING": 5,
    "MULTIPOLYGON": 6,
    "GEOMETRYCOLLECTION": 7,
}

DIMENSIONS_CODE = {None: 0, "Z": 1000, "M": 2000, "ZM": 3000}


def geometry_type_code(wkt):
    if wkt is None:
        return None

    geometry_type, _, dimensions = re.match(r"([A-Z]+)( ([ZM]+)?)?", wkt).groups()
    return GEOMETRY_TYPE_CODE[geometry_type] + DIMENSIONS_CODE[dimensions]


def write_geospatial():
    with open(HERE / "geospatial.yaml") as f:
        examples = yaml.safe_load(f)

    schema = pa.schema({"group": pa.utf8(), "wkt": pa.utf8(), "geometry": ga.wkb()})

    with parquet.ParquetWriter(
        HERE / "geospatial.parquet",
        schema,
        store_schema=False,
        compression="none",
    ) as writer:
        for group_name, geometries_wkt in examples.items():
            # Unfortunately we can't use Shapely to generate the test WKB
            # because of https://github.com/libgeos/geos/issues/888, so we use
            # geoarrow.pyarrow.as_wkb() instead.
            # geometries = shapely.from_wkt(geometries_wkt)
            # wkbs = shapely.to_wkb(geometries, flavor="iso")
            # wkb_array = ga.wkb().wrap_array(pa.array(wkbs, pa.binary()))
            wkt_array = pa.array(geometries_wkt, pa.utf8())

            batch = pa.record_batch(
                {
                    "group": [group_name] * len(geometries_wkt),
                    "wkt": wkt_array,
                    "geometry": ga.as_wkb(wkt_array),
                }
            )

            writer.write_batch(batch)


def write_geospatial_with_nan():
    geometries_wkt = [
        "POINT ZM (10 20 30 40)",
        "POINT ZM (50 60 70 80)",
        "LINESTRING ZM (90 100 110 120, nan nan nan nan, 130 140 150 160)",
    ]

    schema = pa.schema({"group": pa.utf8(), "wkt": pa.utf8(), "geometry": ga.wkb()})

    with parquet.ParquetWriter(
        HERE / "geospatial-with-nan.parquet",
        schema,
        store_schema=False,
        compression="none",
    ) as writer:
        geometries = shapely.from_wkt(geometries_wkt)
        wkbs = shapely.to_wkb(geometries, flavor="iso")
        wkb_array = ga.wkb().wrap_array(pa.array(wkbs, pa.binary()))

        batch = pa.record_batch(
            {
                "group": ["with-nan"] * len(geometries_wkt),
                "wkt": geometries_wkt,
                "geometry": wkb_array,
            }
        )

        writer.write_batch(batch)


def check_geospatial_schema(name):
    file = parquet.ParquetFile(HERE / name)

    col = file.schema.column(2)
    col_dict = json.loads(col.logical_type.to_json())
    col_type = col_dict["Type"]
    if col_type != "Geometry":
        raise ValueError(f"Expected 'Geometry' logical type but got '{col_type}'")


def check_geospatial_values(name):
    tab = parquet.read_table(HERE / name, arrow_extensions_enabled=False)
    geometries_from_wkt = shapely.from_wkt(tab["wkt"])
    geometries_from_wkb = shapely.from_wkb(tab["geometry"])
    shapely.testing.assert_geometries_equal(geometries_from_wkt, geometries_from_wkb)


def iso_type_code(shapely_type_id, has_z, has_m):
    # item was null
    if shapely_type_id < 0:
        return None

    # GEOS type ids are not quite WKB type ids
    iso_type_id = shapely_type_id if shapely_type_id >= 3 else shapely_type_id + 1
    iso_dimensions = (
        3000 if has_z and has_m else 2000 if has_m else 1000 if has_z else 0
    )
    return int(iso_dimensions + iso_type_id)


def calc_stats_shapely(wkts):
    geometries = shapely.from_wkt(wkts)

    # Calculate the list of iso type codes
    any_not_null = any(geom is not None for geom in geometries)
    type_codes = set(geometry_type_code(wkt) for wkt in wkts)

    # Calculate min/max ignoring nan values
    coords = shapely.get_coordinates(geometries, include_z=True, include_m=True)
    coord_mins = [
        None if np.isposinf(x) else float(x)
        for x in np.nanmin(coords, 0, initial=np.inf)
    ]
    coord_maxes = [
        None if np.isneginf(x) else float(x)
        for x in np.nanmax(coords, 0, initial=-np.inf)
    ]

    # Assemble stats in the same format as returned by geo_statistics.to_dict()
    stats = {}
    stats["geospatial_types"] = list(
        sorted(code for code in type_codes if code is not None)
    ) if any_not_null else None
    stats["xmin"], stats["ymin"], stats["zmin"], stats["mmin"] = coord_mins
    stats["xmax"], stats["ymax"], stats["zmax"], stats["mmax"] = coord_maxes

    return stats


def check_batch_statistics(stats, wkts, group):
    if stats is None:
        raise ValueError(f"geo_statistics is missing for group {group}")

    shapely_stats = calc_stats_shapely(wkts)
    file_stats = stats.to_dict()
    if file_stats != shapely_stats:
        raise ValueError(
            f"stats mismatch calculated:\n{shapely_stats}\nvs file\n{file_stats}"
        )


def check_geospatial_statistics(name):
    file = parquet.ParquetFile(HERE / name)
    for i in range(file.num_row_groups):
        batch = file.read_row_group(i)
        column_metadata = file.metadata.row_group(i).column(2)
        group = batch["group"][0].as_py()

        check_batch_statistics(
            column_metadata.geo_statistics, batch["wkt"].to_pylist(), group
        )


if __name__ == "__main__":
    write_geospatial()
    write_geospatial_with_nan()

    check_geospatial_schema("geospatial.parquet")
    check_geospatial_values("geospatial.parquet")
    check_geospatial_statistics("geospatial.parquet")

    check_geospatial_schema("geospatial-with-nan.parquet")
    check_geospatial_statistics("geospatial-with-nan.parquet")
